#   Foma: a finite-state toolkit and library.
#   Copyright © 2008-2015 Mans Hulden
#   Copyright © 2023 Achievement Unlocked Inc., dba CultureFoundry
#
#   This file is part of foma.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# from foma-bindings.fst import Fst
# from foma-bindings import *

# Untested and unclean.
# class MTFSM(Fst):
#     '''Multi-Tape Finite State Machine'''

#     EPSILON_FST = Fst(EPSILON_SYMBOL)

#     def __init__(self, regex=False, num_tapes=2):
#         if isinstance(regex, str):
#             Fst.__init__(self, regex)
#             self.fst_handle = foma_fsm_flatten(foma_fsm_copy(self.fst_handle), foma_fsm_copy(MTFSM.EPSILON_FST.fst_handle))
#             self.num_tapes = num_tapes
#         elif isinstance(regex, Fst):
#             self.fst_handle = foma_fsm_copy(regex.fst_handle)
#             self.regex = None
#             self.num_tapes = num_tapes
#         else:
#             self.fst_handle = None
#             self.regex = None
        
#     def __str__(self):
#         if not self.fst_handle:
#             raise ValueError('FSM not defined')

#         foma_fsm_count(self.fst_handle)
        
#         return f'Name: {self.fst_handle.contents.name}\n\
#                 States: {self.fst_handle.contents.statecount}\n\
#                 Transitions: {self.fst_handle.contents.arccount}\n\
#                 Final States: {self.fst_handle.contents.finalcount}\n\
#                 Deterministic: {self.fst_handle.contents.is_deterministic}\n\
#                 Minimized: {self.fst_handle.contents.is_minimized}\n\
#                 Number of Tapes: {self.num_tapes}'
    
#     def generate(self, word):
#         # TODO variable naming
#         m = self.num_tapes
#         regx = '[{' + word + '}/' + EPSILON_SYMBOL + ' .o. [? 0:?^' + str(m-1) + ']*].l'
#         reg = Fst(regx)
#         gr = Fst()
#         gr.fst_handle = foma_fsm_copy(self.fst_handle)
#         res = MTFSM(reg.intersect(gr), num_tapes = m)
#         return res
    
#     def parse(self, word):
#         # TODO variable naming
#         #[word/□ .o. [0:?^(num_tapes-1) ?]*].l & Grammar ;
#         m = self.num_tapes
#         regx = (u'[{' + word + u'}/□ .o. [0:?^' + str(m-1) + u' ?]*].l')
#         reg = Fst(regx)
#         gr = Fst()
#         gr.fst_handle = foma_fsm_copy(self.fst_handle)
#         res = MTFSM(reg.intersect(gr), num_tapes = m)        
#         return res
    
#     def join(self, other):
#         """Joins two multitape FSMs by composition. E.g.
#             [ a d ]     [ c □ □ ]                  [ a d □ ]
#         A = [ b e ] B = [ d f g ], and A.join(B) = [ b e □ ]
#             [ c □ ]     [ e f g ]                  [ c □ □ ]
#                                                    [ d f g ]
#                                                    [ e f g ] """

#         m = self.num_tapes        
#         n = other.num_tapes
#         pada = Fst('[0:□^' + str(m) +' [0:?^' + str(n-1) + ' - 0:□^' + str(n-1) + '] | ?^' + str(m) + ' 0:?^' + str(n-1) + ']*')
#         padb = Fst('[[0:?^' + str(m-1) + ' - 0:□^' + str(m-1) + ' ] 0:□^' + str(n) + '| 0:?^' + str(m-1) + ' ?^' + str(n) + ']*')
#         extenda = self.compose(pada).lower()
#         extendb = other.compose(padb).lower()
#         flt = Fst('~[?^' + str(m+n-1) +'* [□^' + str(m) + ' ?^' + str(n-1) + ' [?^' + str(m-1) + ' □^' + str(n) + ' |[?^' + str(m-1) +' - □^' + str(m-1) + ' ] □ [?^' + str(n-1) + ' - □^' + str(n-1) + ' ]] | ?^' + str(m-1) + ' □^' + str(n) + ' [□^' + str(m) + ' ?^' + str(n-1) + ' |[?^' + str(m-1) + ' - □^' + str(m-1) + ' ] □ [?^' + str(n-1) + ' - □^' + str(n-1) + ' ]]] ?*]')
#         res = extenda & extendb & flt
#         result = MTFSM(res, m + n - 1)
#         return result

#     def _apply(self, apply_strategy, word=None):
#         if not self.fst_handle:
#             raise ValueError('FST not defined')
#         applyer_handle = foma_apply_init(self.fst_handle)
#         output = apply_strategy(c_void_p(applyer_handle))
#         while True:
#             if output is None:
#                 foma_apply_clear(c_void_p(applyer_handle))
#                 return
#             else:
#                 yield output
#             if word:
#                 output = apply_strategy(c_void_p(applyer_handle), None)
#             else:
#                 output = apply_strategy(c_void_p(applyer_handle))

#     def __iter__(self):
#         return self._mtwords()

#     def __add__(self, other):
#         return self.join(other)

#     def _fmt(self, word):
#         cols = word
#         colchunks = [map(lambda z: len(z), word[x:x+self.num_tapes]) for x in range(0, len(word), self.num_tapes)]
#         col_widths = [max(x) for x in colchunks]
#         format = '  '.join(['%%-%ds' % width for width in col_widths])
#         # string to rows
#         rows = [[word[y] for y in range(x, len(word), self.num_tapes)] for x in range(self.num_tapes)]
#         s = ''
#         for row in rows:
#             #s += format % tuple(row) + '\n'
#             s += ''.join(row) + '\n'
#         return s

#     def _mtwords(self):
#         apply_strategy = foma_apply_upper_words
#         if not self.fst_handle:
#             raise ValueError('FST not defined')
#         applyer_handle = foma_apply_init(self.fst_handle)
#         toksym = '\x07'
#         foma_apply_set_space_symbol(c_void_p(applyer_handle), c_char_p(self.encode(toksym)))
#         output = apply_strategy(c_void_p(applyer_handle))
#         while True:
#             if output is None:
#                 foma_apply_clear(c_void_p(applyer_handle))
#                 return
#             else:
#                 yield self._fmt(self.decode(output)[:-1].split('\x07'))
#             output = apply_strategy(c_void_p(applyer_handle))



#   Foma: a finite-state toolkit and library.
#   Copyright © 2008-2015 Mans Hulden
#   Copyright © 2023 Achievement Unlocked Inc., dba CultureFoundry
#
#   This file is part of foma.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from foma_bindings import *
from sys import maxsize
from ctypes import c_bool

class Fst:
    '''
    This class loads a compiled foma binary into memory, and allows operations on it.

    Here is an example of use:

    ```python
    from foma_bindings import Fst
    fst = Fst.load('/path/to/foma/binary')
    fst.apply_up('fox', flags=Fst.PRINT_FLAGS | Fst.PRINT_SPACES)
    ```

    Constants for printing:
        - PRINT_SPACES
        - PRINT_FLAGS
        - TOKENIZE
        - VERBOSE = TOKENIZE | PRINT_FLAGS

    Word-Accessor Functions:
        - apply_down
        - apply_up
        - lower_words
        - upper_words
        - words

    FST Functions:
        - get_alphabet
        - union
        - intersect
        - minus
        - concatenate
        - compose
        - lower
        - upper
        - flatten
    '''    

    _network_definitions = FstNetworkDefinitions()
    _function_definitions = FstFunctionDefinitions()

    # This is the symbol to split a word on tokens, i.e. "fox+V" becomes ['f','o','x','+V'].
    _TOKENIZE_SYMBOL = b'\x07'

    # Bit flags for different options when printing output.
    # TODO PRINT_SPACES and TOKENIZE do not mesh well together. What to do?
    NO_FLAGS = 0
    PRINT_SPACES = 1
    PRINT_FLAGS = 2
    TOKENIZE = 4
    VERBOSE = TOKENIZE | PRINT_FLAGS

    #region Class and Static Methods
    @classmethod
    def define(cls, definition, name):
        '''Defines an FSM constant; can be supplied regex or existing FSM.
        This link explains the Xerox regex formalism:
        https://github.com/CultureFoundryCA/foma/blob/master/foma/docs/simpleintro.md'''
        # TODO needs a docstring, it's not clear what it means by "defines an FSM constant".
        name = cls.encode(name)
        if isinstance(definition, Fst):
            _ = foma.add_defined(c_void_p(cls._network_definitions.defined_handle), foma_fsm_copy(definition.fst_handle), c_char_p(name))
        elif isinstance(definition, str):
            regex = cls.encode(definition)
            _ = foma.add_defined(c_void_p(cls._network_definitions.defined_handle), foma_fsm_parse_regex(c_char_p(regex), c_void_p(cls._network_definitions.defined_handle), c_void_p(cls._function_definitions.defined_handle)), c_char_p(name))
        else:
            raise ValueError("Expected str, unicode, or FSM")

    @classmethod
    def define_function(cls, prototype, definition):
        """Defines an FSM function."""
        # Prototype is a 2-tuple (name, (arg1name, ..., argname))
        # Definition is regex using prototype variables
        # TODO needs a docstring or needs a better docstring.
        name = cls.encode(prototype[0] + '(')
        if isinstance(definition, str):
            numargs = len(prototype[1])
            for i in range(numargs):
                definition = definition.replace(prototype[1][i], "@ARGUMENT0%i@" % (i+1))
            regex = cls.encode(definition + ';')
            # TODO check this retval
            retval = foma.add_defined_function(c_void_p(cls._function_definitions.deffhandle), c_char_p(name), c_char_p(regex), c_int(numargs))
        else:
            raise ValueError("Expected regex as definition")
        
    @classmethod
    def wordlist(cls, wordlist, minimize=True):
        """Create FSM directly from wordlist.
           Returns a trie-shaped deterministic automaton if not minimized."""
        
        # TODO Fix this naming.
        th = fsm_trie_init()
        for w in wordlist:
            thisword = cls.encode(w)
            fsm_trie_add_word(c_void_p(th), c_char_p(thisword))
        fsm = cls()
        fsm.fst_handle = fsm_trie_done(c_void_p(th))
        if minimize:
            fsm.fst_handle = foma_fsm_minimize(fsm.fst_handle)
        return fsm

    @classmethod
    def load(cls, filename):
        """Load binary FSM from file."""
        fsm = cls()
        fsm.fst_handle = foma_fsm_read_binary_file(c_char_p(Fst.encode(filename)))
        if not fsm.fst_handle:
            raise ValueError("File error.")
        return fsm

    @staticmethod
    def encode(string):
        # TODO clean up this commenting
        # TODO make private
        # type: (Any) -> bytes
        """Makes sure str and unicode are converted."""
        if isinstance(string, str):
            return string.encode('utf8')
        elif isinstance(string, bytes):
            return string
        else:
            return Fst.encode(str(string))

    @staticmethod
    def decode(text):
        # TODO make private
        # TODO needs a docstring or needs a better docstring.
        if text is None:
            return None
        elif isinstance(text, bytes):
            # Assume output is UTF-8 encoded:
            return text.decode('UTF-8')
        else:
            assert isinstance(text, str)
            return text

    #endregion

    #region Private Functions
    def _get_applyer_handle(self, flags):
        '''Initializes a handle to do apply-type functions on. Also sets the output flags on that handle. Flags do not persist once handle is destroyed.'''
        applyer_handle = foma_apply_init(self.fst_handle)

        # Handle the flags to the function.
        if flags & Fst.TOKENIZE:
            foma_apply_set_space_symbol(c_void_p(applyer_handle), c_char_p(Fst._TOKENIZE_SYMBOL))

        if flags & Fst.PRINT_FLAGS:
            foma_apply_set_show_flags(c_void_p(applyer_handle), c_bool(True))

        if flags & Fst.PRINT_SPACES:
            foma_apply_set_print_space(c_void_p(applyer_handle), c_bool(True))
        
        return applyer_handle

    def _apply(self, apply_strategy, word=None, flags=NO_FLAGS):
        '''Takes an application function and executes it on the FST.'''
        if not self.fst_handle:
            raise ValueError('FST not defined')

        applyer_handle = self._get_applyer_handle(flags)

        # TODO why is word set here and not in the while? Can this be cleaned up by always passing the word?
        # Call the apply_function with appropriate word argument.
        if word:
            output = apply_strategy(c_void_p(applyer_handle), c_char_p(self.encode(word)))
        else:
            output = apply_strategy(c_void_p(applyer_handle))

        # Yield results until all FST output for the given input is consumed.
        while True:
            if output is None:
                foma_apply_clear(c_void_p(applyer_handle))
                return
            else:
                if flags & Fst.TOKENIZE:
                    yield [Fst.decode(token) for token in output[:-1].split(Fst._TOKENIZE_SYMBOL)]
                else:
                    yield self.decode(output)

            if word:
                output = apply_strategy(c_void_p(applyer_handle), None)
            else:
                output = apply_strategy(c_void_p(applyer_handle))


    # Functions to call functions through C API.
    def _foma_call_unary(self, unary_func, minimize=True):
        '''Calls a supplied unary function on the FST.'''
        if self.fst_handle:
            handle = unary_func(foma_fsm_copy(self.fst_handle))
            if minimize:
                handle = foma_fsm_minimize(handle)
            return handle
        else:
            raise ValueError('Undefined FST')
        
    def _foma_call_binary(self, other, binary_func, minimize=True):
        '''Calls a supplied binary function on the FST and a provided FST.'''
        if self.fst_handle and other.fst_handle:
            handle = binary_func(foma_fsm_copy(self.fst_handle), foma_fsm_copy(other.fst_handle))
            if minimize:
                handle = foma_fsm_minimize(handle)
            return handle
        else:
            raise ValueError('Undefined FST')

    #endregion

    #region Public Functions
    # Word accessor functions.
    def apply_down(self, word, flags=NO_FLAGS):
        '''Apply a word down the FST. Example: fox+N+Pl -> foxes.'''
        return self._apply(foma_apply_down, word, flags)

    def apply_up(self, word, flags=NO_FLAGS):
        '''Apply a word up the FST. Example: foxes -> fox+N+Pl.'''
        return self._apply(foma_apply_up, word, flags)

    def lower_words(self, flags=NO_FLAGS):
        return self._apply(foma_apply_lower_words, flags=flags)

    def upper_words(self, flags=NO_FLAGS):
        return self._apply(foma_apply_upper_words, flags=flags)

    def words(self, flags=NO_FLAGS):
        return self._apply(foma_apply_words, flags=flags)
        
    # FST operations.
    def get_alphabet(self, minimize=True, flags=NO_FLAGS):
        '''Gets the alphabet of the FST.'''
        sigma = Fst()
        sigma.fst_handle = self._foma_call_unary(foma_fsm_sigma_net, minimize)
        return sigma.words(flags)
    
    # Alias to get_alphabet since both are very different yet valid names for the same function.
    get_sigma = get_alphabet

    def union(self, other, minimize=True):
        '''Returns the union of two FSTs.'''
        fst = Fst()
        fst.fst_handle = self._foma_call_binary(other, foma_fsm_union, minimize)
        return fst

    def intersect(self, other, minimize=True):
        '''Returns the intersection of two FSTs.'''
        fst = Fst()
        fst.fst_handle = self._foma_call_binary(other, foma_fsm_intersect, minimize)
        return fst

    def minus(self, other, minimize=True):
        '''Returns the difference of two FSTs.'''
        fst = Fst()
        fst.fst_handle = self._foma_call_binary(other, foma_fsm_minus, minimize)
        return fst

    def concatenate(self, other, minimize=True):
        '''Returns the concatenation of two FSTs.'''
        fst = Fst()
        fst.fst_handle = self._foma_call_binary(other, foma_fsm_concat, minimize)
        return fst

    def compose(self, other, minimize=True):
        '''Returns the composition of two FSTs.'''
        fst = Fst()
        fst.fst_handle = self._foma_call_binary(other, foma_fsm_compose, minimize)
        return fst
    
    def lower(self, minimize=True):
        fst = Fst()
        fst.fst_handle = self._foma_call_unary(foma_fsm_lower, minimize)
        return fst

    def upper(self, minimize=True):
        fst = Fst()
        fst.fst_handle = self._foma_call_unary(foma_fsm_upper, minimize)
        return fst

    def flatten(self):
        fst = Fst()
        epsilon_fst = Fst(EPSILON_SYMBOL)
        fst.fst_handle = foma_fsm_flatten(foma_fsm_copy(self.fst_handle), foma_fsm_copy(epsilon_fst.fst_handle))
        return fst 
    
    #endregion

    #region Dunders
    def __init__(self, regex=False):
        if regex:
            self.regex = self.encode(regex)
            self.fst_handle = foma_fsm_parse_regex(c_char_p(self.regex), c_void_p(self._network_definitions.defined_handle), c_void_p(self._function_definitions.deffhandle))
            if not self.fst_handle:
                raise ValueError("Syntax error in regex")
        else:
            self.fst_handle = None
        self.get_item_applyer = None

    def __getitem__(self, key):
        if not self.fst_handle:
            raise KeyError('FST not defined')
        if not self.get_item_applyer:
            self.get_item_applyer = foma_apply_init(self.fst_handle)
        result = []
        output = foma_apply_down(c_void_p(self.get_item_applyer), c_char_p(self.encode(key)))
        while True:
            if output is None:
                return result
            else:
                result.append(output)
                output = foma_apply_down(c_void_p(self.get_item_applyer), None)
            
    def __del__(self):
        if self.fst_handle:
            foma_fsm_destroy(self.fst_handle)

    def __str__(self):
        if not self.fst_handle:
            raise ValueError('FSM not defined')
        foma_fsm_count(self.fst_handle)

        return f'Name: {self.fst_handle.contents.name}\n\
                States: {self.fst_handle.contents.statecount}\n\
                Transitions: {self.fst_handle.contents.arccount}\n\
                Final States: {self.fst_handle.contents.finalcount}\n\
                Deterministic: {self.fst_handle.contents.is_deterministic}\n\
                Minimized: {self.fst_handle.contents.is_minimized}\n\
                Arity: {self.fst_handle.contents.arity}\n'

    def __len__(self):
        CYCLIC = -1
        OVERFLOW = -2
        UNKNOWN = -3

        if self.fst_handle:
            if self.fst_handle.contents.pathcount == UNKNOWN:
                self.fst_handle = foma_fsm_topsort(self.fst_handle)
            if self.fst_handle.contents.pathcount == CYCLIC:
                raise ValueError("FSM is cyclic")
            if self.fst_handle.contents.pathcount == OVERFLOW:
                return maxsize
            return self.fst_handle.contents.pathcount
        else:
            raise ValueError("FSM not defined")
        
    def __add__(self, other):
        return self.concat(other)

    def __sub__(self, other):
        return self.minus(other)
    
    def __le__(self, other):
        if self.fst_handle and other.fst_handle:
            return bool(c_int(foma_fsm_isempty(foma_fsm_minimize(foma_fsm_minus(foma_fsm_copy(self.fst_handle),foma_fsm_copy(other.fst_handle))))))
        else:
            raise ValueError('Undefined FST')

    def __lt__(self, other):
        if self.fst_handle and other.fst_handle:
            return (not self.__eq__(other)) and bool(c_int(foma_fsm_isempty(foma_fsm_minimize(foma_fsm_minus(foma_fsm_copy(self.fst_handle),foma_fsm_copy(other.fst_handle))))))
        else:
            raise ValueError('Undefined FST')        
        
    def __or__(self, other):
        return self.union(other)

    def __and__(self, other):
        return self.intersect(other)

    def __eq__(self, other):
        if self.fst_handle and other.fst_handle:
            return bool(c_int(foma_fsm_equivalent(foma_fsm_copy(self.fst_handle), foma_fsm_copy(other.fst_handle))))
        else:
            raise ValueError('Undefined FST')
    
    def __ne__(self, other):
        return not(self.__eq__(other))

    def __contains__(self, word):
        # TODO fix this variable name
        af = self.apply_down(word)
        try:
            # TODO fix this variable name
            i = af.next()
            return True
        except StopIteration:
            return False
                
    def __call__(self, other):
        if isinstance(other, str):
            return Fst("{" + other + "}").compose(self)
        else:
            return other.compose(self)
    
    def __invert__(self):
        fst = Fst()
        fst.fst_handle = self._foma_call_unary(foma_fsm_complement)
        return fst

    def __iter__(self):
        return self._apply(foma_apply_upper_words, word=None, flags=Fst.NO_FLAGS)
    
    #endregion



#   Foma: a finite-state toolkit and library.
#   Copyright © 2008-2015 Mans Hulden
#   Copyright © 2023 Achievement Unlocked Inc., dba CultureFoundry
#
#   This file is part of foma.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# When doing `from foma_bindings import *`, only import foma_bindings and foma_bindings.fst.
all = ['fst']

from ctypes import cdll, c_char, c_int, c_longlong, c_void_p, c_char_p, POINTER, Structure
from ctypes.util import find_library

foma_library_path = find_library('foma')
foma = cdll.LoadLibrary(foma_library_path)

EPSILON_SYMBOL = '□'

class FstStruct(Structure):
    _fields_ = [
        ("name", c_char * 40),
        ("arity", c_int),
        ("arccount", c_int),
        ("statecount", c_int),
        ("linecount", c_int),
        ("finalcount", c_int),
        ("pathcount", c_longlong),
        ("is_deterministic", c_int),
        ("is_pruned", c_int),
        ("is_minimized", c_int),
        ("is_epsilon_free", c_int),
        ("is_loop_free", c_int),
        ("is_completed", c_int),
        ("arcs_sorted_in", c_int),
        ("arcs_sorted_out", c_int),
        ("fsm_state", c_void_p),
        ("sigma", c_void_p),
        ("medlookup", c_void_p)
    ]

# FST functions.
foma_fsm_complement = foma.fsm_complement
foma_fsm_complement.restype = POINTER(FstStruct)
foma_fsm_compose = foma.fsm_compose
foma_fsm_compose.restype = POINTER(FstStruct)
foma_fsm_concat = foma.fsm_concat
foma_fsm_concat.restype = POINTER(FstStruct)
foma_fsm_copy = foma.fsm_copy
foma_fsm_copy.restype = POINTER(FstStruct)
foma_fsm_count = foma.fsm_count
foma_fsm_destroy = foma.fsm_destroy
foma_fsm_equivalent = foma.fsm_equivalent
foma_fsm_equivalent.restype = c_int
foma_fsm_flatten = foma.fsm_flatten
foma_fsm_flatten.restype = POINTER(FstStruct)
foma_fsm_intersect = foma.fsm_intersect
foma_fsm_intersect.restype = POINTER(FstStruct)
foma_fsm_isempty = foma.fsm_isempty
foma_fsm_isempty.restype = c_int
foma_fsm_lower = foma.fsm_lower
foma_fsm_lower.restype = POINTER(FstStruct)
foma_fsm_minimize = foma.fsm_minimize
foma_fsm_minimize.restype = POINTER(FstStruct)
foma_fsm_minus = foma.fsm_minus
foma_fsm_minus.restype = POINTER(FstStruct)
foma_fsm_parse_regex = foma.fsm_parse_regex
foma_fsm_parse_regex.restype = POINTER(FstStruct)
foma_fsm_read_binary_file = foma.fsm_read_binary_file
foma_fsm_read_binary_file.restype = POINTER(FstStruct)
foma_fsm_sigma_net = foma.fsm_sigma_net
foma_fsm_sigma_net.restype = POINTER(FstStruct)
foma_fsm_topsort = foma.fsm_topsort
foma_fsm_topsort.restype = POINTER(FstStruct)
foma_fsm_union = foma.fsm_union
foma_fsm_union.restype = POINTER(FstStruct)
foma_fsm_upper = foma.fsm_upper
foma_fsm_upper.restype = POINTER(FstStruct)

# Apply functions.
foma_apply_clear = foma.apply_clear
foma_apply_down = foma.apply_down
foma_apply_down.restype = c_char_p
foma_apply_init = foma.apply_init
foma_apply_init.restype = c_void_p
foma_apply_lower_words = foma.apply_lower_words
foma_apply_lower_words.restype = c_char_p
foma_apply_up = foma.apply_up
foma_apply_up.restype = c_char_p
foma_apply_upper_words = foma.apply_upper_words
foma_apply_upper_words.restype = c_char_p
foma_apply_words = foma.apply_words
foma_apply_words.restype = c_char_p

# Set functions.
foma_apply_set_print_space = foma.apply_set_print_space
foma_apply_set_show_flags = foma.apply_set_show_flags
foma_apply_set_space_symbol = foma.apply_set_space_symbol

# Define functions.
defined_functions_init = foma.defined_functions_init
defined_functions_init.restype = c_void_p
defined_networks_init = foma.defined_networks_init
defined_networks_init.restype = c_void_p
foma_add_defined = foma.add_defined
foma_add_defined.restype = c_int
foma_add_defined_function = foma.add_defined_function
foma_add_defined_function.restype = c_int

# Trie functions.
fsm_trie_add_word = foma.fsm_trie_add_word
fsm_trie_done = foma.fsm_trie_done
fsm_trie_done.restype = POINTER(FstStruct)
fsm_trie_init = foma.fsm_trie_init
fsm_trie_init.restype = c_void_p

# TODO What is this for?
class FstNetworkDefinitions:
    def __init__(self):
        self.defined_handle = defined_networks_init(None)

# TODO What is this for?
class FstFunctionDefinitions:
    def __init__(self):
        self.defined_handle = defined_functions_init(None)



#   Copyright © 2023 Achievement Unlocked Inc., dba CultureFoundry
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from setuptools import setup, find_packages

setup(
    name='foma_bindings',
    version='0.1.3',
    description='Python bindings for the foma finite-state technology suite',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/CultureFoundryCA/foma_bindings',
    author='CultureFoundry',
    author_email='info@culturefoundrystudios.com',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3'
    ],
    keywords=['foma', 'xfst', 'hfst', 'fst', 'fsm', 'mtfst', 'mtfsm'],
    packages=find_packages(include=['foma_bindings', 'foma_bindings.fst', 'foma_bindings.mtfst']),
    python_requires='>=3.10, <4',
)
